local addonName, BaseAP = ...
_G.BaseAP = BaseAP
setmetatable(BaseAP, {__index = function(self, index) return _G[index] end})
setfenv(1, BaseAP)

local _G = _G
local print = print
local tostring = tostring
local tonumber = tonumber
local select = select
local pairs = pairs
local unpack = unpack
local strlen = strlen
local strfind = strfind
local strsub = strsub
local strmatch = strmatch
local gsub = gsub
local gmatch = gmatch
local format = format
local floor = floor
local tconcat = table.concat
local tinsert = tinsert
local wipe = wipe
local GetTime = GetTime
local BreakUpLargeNumbers = BreakUpLargeNumbers

debug = false
eventFrame = CreateFrame("Frame")
LAD = LibStub("LibArtifactData-1.0")
knowlMult = 1 -- knowledge multiplier
transformCache = {}
transformCacheNumEntries = 0

-- Utility
function debugPrint(...)
	if debug then
		print("|cff00ffff"..addonName..":|r", ...)
	end
end

unitTable = {
	["million"]  = 1000000,
	["billion"]  = 1000000000,
	["trillion"] = 1000000000000,
}

-- actually maps from (AK+1) to multiplier
akMults = {
	1, 1.25, 1.5, 1.9, 2.4, 3,
	3.75, 4.75, 6, 7.5, 9.5,
	12, 15, 18.75, 23.5, 29.5,
	37, 46.5, 58, 73, 91,
	114, 143, 179, 224, 250,
	1001, 1301, 1701, 2201, 2901,
	3801, 4901, 6401, 8301, 10801,
	14001, 18201, 23701, 30801, 40001,
	160001, 208001, 270401, 351501, 457001,
	594001, 772501, 1004001, 1305001, 1696501,
	2205501, 2867501, 3727501, 4846001, 6300001,
}

-- Get player's character name and realm
function getCharFullName()
	local realm = GetRealmName()
	if not realm then
		return nil
	end
	local name = UnitName("player")
	if not name then
		return nil
	end
	return name.."-"..gsub(realm, "%s+", "")
end

-- Per-character caching of AK multiplier
function getCachedKnowlMult()
	local fullName = getCharFullName()
	-- debugPrint("getCachedKnowlMult", fullName, BaseAPSaved.knowlMults[fullName])
	return BaseAPSaved.knowlMults[fullName]
end

function setCachedKnowlMult(mult)
	local fullName = getCharFullName()
	-- debugPrint("setCachedKnowlMult", "fullName:", fullName, "mult:", mult)
	if fullName then
		BaseAPSaved.knowlMults[fullName] = mult
	end
end

-- LibArtifactData interfacing
function BaseAP:ARTIFACT_KNOWLEDGE_CHANGED(message, level, mult)
	knowlMult = mult
	setCachedKnowlMult(mult)
	wipe(transformCache)
end

-- Returns: amount, isExact
-- Example: parseAmountStr("100,000") => 100000, true
--          parseAmountStr("2.3 million") => 2300000, false
function parseAmountStr(amountStr)
	local k = strfind(amountStr, " ", 1, true)
	if k then
		local units = strsub(amountStr, k + 1)
		units = unitTable[units]
		if not units then return end
		local qty = tonumber(strsub(amountStr, 1, k - 1))
		if not qty then return end
		return qty * units, false
	else
		amountStr = gsub(amountStr, ",", "")
		return tonumber(amountStr), true
	end
end

-- Implement "advanced scanning" technique to determine base AP:
-- Given a link, construct a "base" version of the link, load it into
-- our scan tooltip, then scan the tooltip.

-- To scan, we call SetHyperlink(). It seems that if we naively do a scan afterwards,
-- it sometimes doesn't work for some reason, NumLines() returns 0. However,
-- if we do the scan inside an OnTooltipSetItem hook, it seems to work.
-- Hence this complicated setup.

function getTooltipAP(tooltip)
	local name = tooltip:GetName()
	for i = 3, tooltip:NumLines() do
		local textFrame = _G[name.."TextLeft"..i]
		if not textFrame then break end
		local text = textFrame:GetText()
		if text then
			local amountStr = strmatch(text, "^Use: Grants (.*) Artifact Power")
			if amountStr then
				return parseAmountStr(amountStr)
			end
		end
	end
end

BaseAPScanTooltip_AP = nil
function BaseAPScanTooltip_OnTooltipSetItem(tooltip)
	BaseAPScanTooltip_AP = getTooltipAP(tooltip)
end

function scanAP(link)
	BaseAPScanTooltip_AP = nil
	BaseAPScanTooltip:SetHyperlink(link)
	return BaseAPScanTooltip_AP
end

-- Link parsing
function destructLink(link, tokens)
	tokens = tokens or {}
	wipe(tokens)

	local linkColor, linkMiddle, name = strmatch(link, "^|c([^|]*)|Hitem([^|]*)|h%[([^|]*)%]|h|r$")
	if not name then -- failed to match
		return
	end
	for token in gmatch(linkMiddle, ":(%d*)") do
		tinsert(tokens, token)
	end
	return tokens, linkColor, name
end

function constructLink(tokens, linkColor, name)
	return "|c"..linkColor.."|Hitem:"..tconcat(tokens, ":").."|h["..name.."]|h|r"
end

-- Construct the "base" version of the link, then call scanAP.
-- Interpret the scan as having failed iff the first return value (baseAP) is nil.
scanTokens = {} -- reuse table as optimization
function scanBaseAPAK_(link)
	local tokens, linkColor, name = destructLink(link, scanTokens)
	if not tokens then return end

	local numBonusIDs = tonumber(tokens[13]) or 0
	local i = 14 + numBonusIDs
	local itemAK = nil
	if tokens[11] == "" then -- item shown at current AK
		tinsert(tokens, i, "1")
	else -- item has specific AK
		itemAK = tonumber(tokens[i])
		tokens[i] = "1"
	end
	tokens[11] = "8388608"

	local baseLink = constructLink(tokens, linkColor, name)
	local baseAP = scanAP(baseLink)
	return baseAP, itemAK
end

-- Caching for scanBaseAPAK_ (minimize amount of tooltip scanning)
-- If scan succeeds, it never retries. If it fails, it can retry after a cooldown.
baseAPAKCache = {}
function scanBaseAPAK(link)
	if not link then return end
	local baseAP, itemAK, timestamp
	local cacheEntry = baseAPAKCache[link]
	if cacheEntry then
		baseAP, itemAK, timestamp = unpack(cacheEntry)
		if baseAP or GetTime() - timestamp < 2 then
			return baseAP, itemAK
		end
	end
	baseAP, itemAK = scanBaseAPAK_(link)
	baseAPAKCache[link] = {baseAP, itemAK, GetTime()}
	return baseAP, itemAK
end

-- End of "advanced scanning" section.

-- Transforming tooltip text
function transformAPText(text, curMult, link)
	-- If match fails or function returns nil, `text` is returned unmodified. XXX
	local result, numMatches = gsub(text, "^Use: Grants (.*) Artifact Power", function(amountStr)
		if strfind(text, "not affected by Artifact Knowledge", 50, true) then return end
		amountStr = gsub(amountStr, " %([^)]*%)", "") -- for SetLFGDungeonReward workaround
		local amount, isExact = parseAmountStr(amountStr)
		local color = "ffffffff"
		if not amount then return end
		if isExact then
			amount = amount / curMult
		else
			local baseAP, itemAK = scanBaseAPAK(link)
			if baseAP then
				if itemAK then -- item has specific AK
					if akMults[itemAK] then
						amount = baseAP * akMults[itemAK] / curMult
					else -- don't have AK multiplier information, fall back
						amount = amount / curMult
						color = "ffffff88"
					end
				else -- item shown at current AK
					amount = baseAP
				end
			else -- fall back
				amount = amount / curMult
				color = "ffffff88"
			end
		end
		amount = 5 * floor(0.2 * amount + 0.5) -- round to multiple of 5
		return format("Use: Grants %s (|c%s%s|r) Artifact Power", amountStr, color, BreakUpLargeNumbers(amount))
	end)
	return numMatches > 0 and result or false
end

function modifyTooltip(tooltip, link)
	if knowlMult == 1 then return end
	if transformCacheNumEntries > 50 then
		wipe(transformCache)
		transformCacheNumEntries = 0
	end

	local _
	if not link then
		_, link = tooltip:GetItem()
	end
	local name = tooltip:GetName()

	for i = 3, tooltip:NumLines() do
		local textFrame = _G[name.."TextLeft"..i]
		local text = textFrame:GetText()
		if text then
			if transformCache[text] == false then -- leave unchanged
			else
				local combinedKey = text..tostring(link)
				local cacheEntry = transformCache[combinedKey]
				if cacheEntry == nil then
					cacheEntry = transformAPText(text, knowlMult, link)
					if cacheEntry == false then
						transformCache[text] = false
						transformCacheNumEntries = transformCacheNumEntries + 1
					else
						transformCache[combinedKey] = cacheEntry
						transformCacheNumEntries = transformCacheNumEntries + 1
						textFrame:SetText(cacheEntry)
					end
				else
					textFrame:SetText(cacheEntry)
				end
			end
		end
	end
end

-- Workaround for LFG reward tooltips having a malformed link returned by GetItem().
-- We hook the SetLFGDungeonReward function to get the info. This hook gets called
-- after the usual OnTooltipSetItem, so we end up modifying the tooltip *twice*.

function GameTooltip_SetLFGDungeonShortageReward(self, dungeonID, shortageIndex, lootIndex)
	local link = GetLFGDungeonShortageRewardLink(dungeonID, shortageIndex, lootIndex)
	-- debugPrint("GameTooltip_SetLFGDungeonShortageReward", link)
	modifyTooltip(self, link)
end

function GameTooltip_SetLFGDungeonReward(self, dungeonID, lootIndex)
	local link = GetLFGDungeonRewardLink(dungeonID, lootIndex)
	-- debugPrint("GameTooltip_SetLFGDungeonReward", link)
	modifyTooltip(self, link)
end

function init()
	debugPrint("Init started")

	-- Initialize the SavedVariable
	BaseAPSaved = BaseAPSaved or {}
	BaseAPSaved.knowlMults = BaseAPSaved.knowlMults or {}

	-- Get per-character cached AK multiplier if available
	knowlMult = getCachedKnowlMult() or select(2, LAD:GetArtifactKnowledge()) or 1
	debugPrint("Initial guess is", knowlMult)

	-- Initialize the scan tooltip
	CreateFrame("GameTooltip", "BaseAPScanTooltip", nil, "GameTooltipTemplate")
	BaseAPScanTooltip:SetOwner(UIParent, "ANCHOR_NONE")
	BaseAPScanTooltip:HookScript("OnTooltipSetItem", BaseAPScanTooltip_OnTooltipSetItem)

	-- Tell LibArtifactData to inform us when AK multiplier changes or becomes available
	LAD.RegisterCallback(BaseAP, "ARTIFACT_KNOWLEDGE_CHANGED")

	-- Watch for tooltip init, then hook
	local hooksPending = {
		["GameTooltip"] = true,
		["WorldMapTooltipTooltip"] = true,
		["PVPRewardTooltipTooltip"] = true,
		["ItemRefTooltip"] = true,
		["EncounterJournalTooltipItem1Tooltip"] = true,
		["ReputationParagonTooltipTooltip"] = true,
	}
	local ticker
	ticker = C_Timer.NewTicker(1, function()
		local allDone = true
		for name, _ in pairs(hooksPending) do
			allDone = false
			local tooltip = _G[name]
			if tooltip then
				modifyTooltip(tooltip)
				tooltip:HookScript("OnTooltipSetItem", modifyTooltip)
				debugPrint("Hooked", name)
				hooksPending[name] = nil
			end
		end
		if allDone then
			ticker:Cancel()
		end
	end)

	-- Special hook for LFG dungeon rewards
	hooksecurefunc(GameTooltip, "SetLFGDungeonShortageReward", GameTooltip_SetLFGDungeonShortageReward)
	hooksecurefunc(GameTooltip, "SetLFGDungeonReward", GameTooltip_SetLFGDungeonReward)
end

eventFrame:RegisterEvent("ADDON_LOADED")
eventFrame:SetScript("OnEvent", function(self, event, arg1, ...)
	if event == "ADDON_LOADED" and addonName == arg1 then
		eventFrame:UnregisterEvent("ADDON_LOADED")
		init()
	end
end)
